########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############


def package_completer(env, prefix, **_):
    nodes = env.storage.get_nodes()
    suffix_len = len('-package')
    return (n.id[:-suffix_len] for n in nodes
            if n.type == 'python_package' and
            n.id.startswith(prefix))


def repo_completer(env, prefix, **_):
    nodes = env.storage.get_nodes()
    suffix_len = len('-repo')
    return (n.id[:-suffix_len] for n in nodes
            if n.type == 'git_repo' and
            n.id.startswith(prefix))


def active_feature_repo_completer(env, prefix, **_):
    import json
    with open(env.storage._payload_path) as f:
        active_feature = json.load(f).get('active_feature')
    if not active_feature:
        return []
    features = _load_features_file(env)
    active_feature = features.get(active_feature)
    if not active_feature:
        return []
    repos = active_feature.get('repos', [])
    return (repo for repo in repo_completer(env, prefix)
            if repo in repos)


def features_completer(env, prefix, **_):
    features = _load_features_file(env)
    return [k for k in features if k.startswith(prefix)]


def branches_completer(env, prefix, **_):
    versions_prefix = '::'
    if prefix.startswith(versions_prefix):
        from path import path
        from sh import git
        import json
        with open(env.storage._payload_path) as f:
            versions_repo_location = path(
                json.load(f)['versions_repo_location'])
        with versions_repo_location:
            result = git('show-ref').stdout.split('\n')
            result = ['{}{}'.format(versions_prefix,
                                    r.strip().rsplit('/', 1)[-1])
                      for r in result]
            return [r for r in result if r.startswith(prefix)]
    else:
        result = []
        if 'default'.startswith(prefix):
            result.append('default')
        return result


def _load_features_file(env):
    features_file = env.plan['inputs'].get('features_file')
    if not features_file:
        return []
    import os
    features_file = os.path.expanduser(features_file)
    if not os.path.exists(features_file):
        return []
    import yaml
    with open(features_file) as f:
        features = yaml.safe_load(f) or {}
    return features



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import os
import json

import sh
from path import path

from cloudify import ctx
from cloudify.decorators import operation
from cloudify import exceptions

from common import bake


@operation
def makedirs(location, **_):
    location = os.path.expanduser(location)
    if not os.path.isdir(location):
        os.makedirs(location)


@operation
def nose_run(virtualenv_location, test_path, **_):
    nose = bake(sh.Command(os.path.join(virtualenv_location, 'bin',
                                        'nosetests')))
    try:
        nose(test_path,
             nocapture=True,
             nologcapture=True,
             verbose=True).wait()
    except:
        raise exceptions.NonRecoverableError()


@operation
def configure_docs_getcloudify_source(docs_getcloudify_repo_location, **_):
    repo_location = path(
        ctx.source.instance.runtime_properties['repo_location'])
    dev_directory = repo_location / 'dev'
    config_path = dev_directory / 'config.json'

    if not dev_directory.exists():
        dev_directory.mkdir()

    config_path.write_text(json.dumps({
        'content': {
            'root': docs_getcloudify_repo_location
        }
    }))



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############


def after_env_create(loader, **kwargs):
    features_yaml = loader.user_config.storage_dir / 'features.yaml'
    features_yaml.touch()


def before_init(blueprint, inputs, **kwargs):
    node_templates = blueprint['node_templates']
    repos = inputs.pop('repos', {})
    for repo_name, repo in repos.items():
        repo_node_template_name = '{}-repo'.format(repo_name)
        repo_type = repo.get('type', 'misc')
        properties = repo.get('properties', {})
        properties.update({
            'name': repo_name,
            'repo_type': repo_type
        })
        node_templates[repo_node_template_name] = {
            'type': 'git_repo',
            'properties': properties,
            'relationships': [{
                'target': 'repos_dir',
                'type': 'cloudify.relationships.depends_on'
            }]
        }
        python = repo.get('python', repo_type != 'versions')
        if not python:
            continue
        elif python is True:
            python = [{
                'name': repo_name
            }]
        elif isinstance(python, dict):
            python = [python]
        for python_package in python:
            package_name = python_package.get('name', repo_name)
            package_node_template_name = '{}-package'.format(package_name)
            node_templates[package_node_template_name] = {
                'type': 'python_package',
                'properties': {
                    'name': package_name,
                    'path': python_package.get('path', ''),
                    'test': python_package.get('test', '')
                },
                'relationships': [
                    {'target': repo_node_template_name,
                     'type': 'package_depends_on_repo'},
                    {'target': 'virtualenv',
                     'type': 'package_depends_on_virtualenv'}
                ]
            }
            dependencies = python_package.get('dependencies', [])
            if (repo.get('type') == 'plugin' and
                    'cloudify-plugins-common' not in dependencies):
                dependencies.append('cloudify-plugins-common')
            for dependency in dependencies:
                relationships = node_templates[package_node_template_name][
                    'relationships']
                relationships.append({
                    'target': '{}-package'.format(dependency),
                    'type': 'package_depends_on_package'
                })
    package_installer = node_templates['package_installer']
    relationships = package_installer['relationships']
    for node_template_name, node_template in node_templates.items():
        if node_template['type'] == 'python_package':
            relationships.append({
                'target': node_template_name,
                'type': 'cloudify.relationships.depends_on'
            })
    idea_project = node_templates['idea_project']
    relationships = idea_project.get('relationships', [])
    idea_project['relationships'] = relationships
    for node_template_name, node_template in node_templates.items():
        if node_template['type'] in ['python_package', 'git_repo']:
            relationships.append({
                'target': node_template_name,
                'type': 'cloudify.relationships.depends_on'
            })
    docs = node_templates.get('docs.getcloudify.org-repo')
    docs_site = node_templates.get('docs.getcloudify.org-site-repo')
    if docs and docs_site:
        relationships = docs_site['relationships']
        relationships.append({
            'target': 'docs.getcloudify.org-repo',
            'type': 'docs_site_depends_on_docs'
        })



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import os
import sys
import tempfile

import sh
import networkx
from path import path

from cloudify import ctx
from cloudify.decorators import operation

from common import bake


@operation
def configure_python_package_path(repo_location, base_package_path, **_):
    package_path = os.path.join(repo_location, base_package_path)
    ctx.source.instance.runtime_properties['package_path'] = package_path


@operation
def configure_virtualenv_location(virtualenv_location, **_):
    ctx.source.instance.runtime_properties[
        'virtualenv_location'] = virtualenv_location


@operation
def configure_package_dependency(target_package_path, **_):
    _add_requirement(target_package_path, ctx.source.instance)


@operation
def add_self_requirement(package_path, **_):
    _add_requirement(package_path, ctx.instance)


@operation
def add_requirements(requirements, **_):
    ctx.instance.runtime_properties['requirements'] = requirements or []


def _add_requirement(package_path, instance):
    requirements = instance.runtime_properties.get('requirements', [])
    requirement = '-e {}'.format(package_path)
    if requirement not in requirements:
        requirements.append(requirement)
    instance.runtime_properties['requirements'] = requirements


@operation
def install(virtualenv_location, package_path, **_):
    pip = bake(sh.Command(os.path.join(virtualenv_location, 'bin', 'pip')))
    pip.install(
        c=os.path.join(virtualenv_location, 'constraints.txt'),
        e=package_path).wait()


@operation
def install_packages(**_):
    storage = ctx._endpoint.storage
    package_node_instance_ids = ctx.capabilities.get_all().keys()
    package_node_instances = [storage.get_node_instance(instance_id)
                              for instance_id in package_node_instance_ids]
    package_node_instances = [
        instance for instance in package_node_instances
        if 'virtualenv_location' in instance.runtime_properties]
    virtualenvs = set((instance.runtime_properties['virtualenv_location']
                       for instance in package_node_instances))

    for virtualenv_location in virtualenvs:
        virtualenv_package_instances = [
            instance for instance in package_node_instances
            if instance.runtime_properties[
                'virtualenv_location'] == virtualenv_location]
        graph = networkx.DiGraph()
        for instance in virtualenv_package_instances:
            graph.add_node(instance.id)
            for relationship in instance.relationships:
                graph.add_edge(instance.id, relationship['target_id'])
        topological_sort = networkx.topological_sort(graph.reverse())
        requirements = []
        for instance_id in topological_sort:
            instance = storage.get_node_instance(instance_id)
            for requirement in instance.runtime_properties.get('requirements',
                                                               []):
                if requirement not in requirements:
                    requirements.append(requirement)
        requirements_file = tempfile.mktemp(prefix='requirements-',
                                            suffix='.txt')
        if not requirements:
            return
        with open(requirements_file, 'w') as f:
            f.write('\n'.join(requirements))
        pip = bake(sh.Command(os.path.join(virtualenv_location, 'bin', 'pip')))
        pip.install(
            requirement=requirements_file,
            c=os.path.join(virtualenv_location, 'constraints.txt')).wait()


@operation
def configure_virtualenv(
        virtualenv_location,
        constraints,
        postactivate_resource_path,
        git_retag_cloudify_resource_path,
        register_python_argcomplete,
        **_):

    # constraints.txt
    constraints = constraints or []
    constraints_path = os.path.join(
        ctx.instance.runtime_properties['virtualenv_location'],
        'constraints.txt')
    with open(constraints_path, 'w') as f:
        f.write('\n'.join(constraints))

    virtualenv_bin = path(virtualenv_location) / 'bin'

    # clue symlink
    clue_source_path = path(sys.executable).dirname() / 'clue'
    clue_target_path = virtualenv_bin / 'clue'
    if not clue_target_path.exists():
        os.symlink(clue_source_path, clue_target_path)

    # postactivate
    variables = dict(register_python_argcomplete=register_python_argcomplete)
    ctx.download_resource_and_render(
        postactivate_resource_path,
        target_path=virtualenv_bin / 'postactivate',
        template_variables=variables)

    # git-retag-cloudify
    retag_cloudify_target_path = virtualenv_bin / 'git-retag-cloudify'
    ctx.download_resource_and_render(
        git_retag_cloudify_resource_path,
        template_variables={'sys_executable': sys.executable},
        target_path=retag_cloudify_target_path)
    os.chmod(retag_cloudify_target_path, 0755)



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import json
import os
import sys

import sh
import yaml
from path import path

from cloudify import ctx
from cloudify.workflows import ctx as workflow_ctx
from cloudify import exceptions
from cloudify.decorators import operation
from cloudify.decorators import workflow

from common import bake


class GitRepo(object):

    def clone(self):
        git = bake(sh.git)
        self.runtime_properties['repo_location'] = str(self.repo_location)
        self.git_version = sh.git(version=True).stdout.strip()
        if self.repo_location.isdir():
            return
        if self.clone_method == 'https':
            clone_url = 'https://github.com/{}/{}.git'.format(
                self.organization, self.name)
        elif self.clone_method == 'ssh':
            clone_url = 'git@github.com:{}/{}.git'.format(
                self.organization, self.name)
        else:
            raise exceptions.NonRecoverableError(
                'Illegal clone method: {0}'.format(self.clone_method))
        git.clone(clone_url, self.repo_location, '-b', self.branch).wait()

    def configure(self):
        # configure commit-msg hook
        hook_path = self.repo_location / '.git' / 'hooks' / 'commit-msg'
        ctx.download_resource_and_render(
            'resources/commit-msg',
            template_variables={'sys_executable': sys.executable},
            target_path=hook_path)
        os.chmod(hook_path, 0o755)
        # git config
        for key, value in self.git_config.items():
            self.git.config(key, value).wait()
        if self.type == 'versions':
            self.versions_repo_location = self.repo_location

    def pull(self):
        try:
            self.git_output('rev-parse', '@{u}')
        except sh.ErrorReturnCode:
            ctx.logger.info('No upstream defined. Skipping pull.')
            return
        kwargs = {'prune': True}
        if self.git_version >= 'git version 2.0.0':
            kwargs['tags'] = True
        self.git.pull(**kwargs).wait()

    def status(self, active):
        if active and not self.active_feature.branch:
            return
        for git_prompt_path in self.git_prompt_paths:
            if git_prompt_path.expanduser().exists():
                break
        else:
            git_prompt_path = None
        if git_prompt_path:
            script_path = ctx.download_resource_and_render(
                'resources/git-branch-state.sh', template_variables={
                    'git_prompt_path': git_prompt_path,
                    'repo_location': self.repo_location})
            try:
                os.chmod(script_path, 0o755)
                script = sh.Command(script_path)
                branch_state = script().stdout.strip()
            finally:
                os.remove(script_path)
        else:
            branch_state = self.current_branch
        ctx.logger.info(branch_state)
        self.git.status(s=True).wait()

    def checkout(self, branch):
        versions_prefix = '::'
        default_branch = self.branch
        active_feature = self.active_feature
        if branch.startswith(versions_prefix):
            versions_branch = branch[len(versions_prefix):]
            components = self._read_versions_file(versions_branch)
            if self.name in components:
                branch = components[self.name]
            elif self.type in ['core', 'versions']:
                branch = versions_branch
            else:
                branch = default_branch
        elif active_feature.name == branch:
            branch = active_feature.branch
            base_branch = active_feature.base
            if not branch:
                if base_branch.startswith(versions_prefix):
                    versions_branch = base_branch[len(versions_prefix):]
                    components = self._read_versions_file(versions_branch)
                    if self.name in components:
                        branch = components[self.name]
                    elif self.type in ['core', 'versions']:
                        branch = versions_branch
                    else:
                        branch = default_branch
                else:
                    if self.type in ['core', 'versions']:
                        branch = base_branch
                    else:
                        branch = default_branch
        elif branch == 'default':
            branch = default_branch
        elif self.type == 'misc':
            return
        elif self.type not in ['core', 'plugin', 'versions']:
            raise exceptions.NonRecoverableError('Unhandled repo type: {}'
                                                 .format(self.type))
        elif not branch:
            raise exceptions.NonRecoverableError('Branch is not defined')
        elif branch.startswith('.'):
            branch = self._fix_branch_name(branch)
        try:
            if self.current_branch != branch:
                self.git.checkout(branch).wait()
        except sh.ErrorReturnCode:
            ctx.logger.error('Could not checkout branch {0}'.format(branch))

    def reset(self, hard, origin):
        if not self.validate_active_feature():
            return
        self.git.reset.bake(hard=hard)(
            '{0}/{1}'.format(origin, self.current_branch)).wait()

    def rebase(self):
        if not self.validate_active_feature():
            return
        base = self.active_feature.base
        try:
            base = self._fix_branch_name(base)
            self.git.rebase(base).wait()
        except Exception as e:
            ctx.logger.error('Failed rebase, aborting: {0}'.format(e))
            try:
                self.git.rebase(abort=True).wait()
            except:
                pass

    def squash(self):
        if not self.validate_active_feature():
            return
        if self.git_version < 'git version 1.9':
            ctx.logger.warn('git version >= 1.9 is required for squash.')
            return
        git = self.git_output
        base = self.active_feature.base
        base = self._fix_branch_name(base)
        merge_base = git.bake('merge-base', fork_point=True)(
                base).stdout.strip()
        commits = git.bake(
            'rev-list',
            ancestry_path=True,
        )('{0}..HEAD'.format(merge_base)).stdout.strip().split('\n')
        commits = [c.strip() for c in commits]
        if len(commits) == 1:
            ctx.logger.info('Single commit found. Skipping squash.')
            return
        commit_message_sha = commits[-1]
        commit_message = git.log(commit_message_sha,
                                 format='%B', n=1).stdout.strip()
        git = self.git
        ctx.logger.info(
            'Squashing with merge_base: {0} and commit message: {1}'
            .format(merge_base, commit_message))
        git.reset(merge_base, soft=True).wait()
        git.commit(message=commit_message).wait()

    def diff(self, revision_range, cached, active):
        if active and not self.active_feature.branch:
            return
        if self.type not in ['core', 'plugin']:
            return

        args = []
        kwargs = {
            'cached': cached
        }

        if revision_range:
            split_by = '..'
            split_range = revision_range.split(split_by)
            if len(split_range) != 2:
                ctx.logger.error(
                    'Invalid range supplied: {0}'.format(revision_range))
                return
            left = self._fix_branch_name(split_range[0])
            right = self._fix_branch_name(split_range[1])
            diff_range = '{0}{1}{2}'.format(left, split_by, right)
            args.append(diff_range)

        try:
            self.git.diff(*args, **kwargs).wait()
        except sh.ErrorReturnCode:
            ctx.logger.error('{0}, {1} diff failed'.format(args, kwargs))

    def create_branch(self, branch, base):
        if not self._branch_exists(branch):
            self.git.branch(branch, base).wait()

    def delete_branch(self, branch, force):
        if self._branch_exists(branch, local_only=True):
            if self.current_branch == branch:
                self.git.checkout(self.branch).wait()
            try:
                delete_flag = '-D' if force else '-d'
                self.git.branch(delete_flag, branch).wait()
            except sh.ErrorReturnCode:
                ctx.logger.error('Failed deleting branch {0}.'.format(branch))

    def branch_exists(self, branch):
        return self._branch_exists(branch)

    @property
    def location(self):
        return path(self.properties['location']).expanduser()

    @property
    def name(self):
        return self.properties['name']

    @property
    def repo_location(self):
        return self.location / self.name

    @property
    def organization(self):
        return self.properties['organization']

    @property
    def branch(self):
        return self.properties['branch']

    @property
    def clone_method(self):
        return self.properties['clone_method']

    @property
    def type(self):
        return self.properties['repo_type']

    @property
    def git_config(self):
        return self.properties['git_config']

    @property
    def git_prompt_paths(self):
        return [path(p) for p in self.properties['git_prompt_paths']]

    @property
    def git_version(self):
        return self.runtime_properties['git_version']

    @git_version.setter
    def git_version(self, value):
        self.runtime_properties['git_version'] = value

    @property
    def current_branch(self):
        return self.git_output('rev-parse', '--abbrev-ref',
                               'HEAD').stdout.strip()

    @property
    def git(self):
        return self._git(log_out=True)

    @property
    def git_output(self):
        return self._git(log_out=False)

    @property
    def properties(self):
        return ctx.node.properties

    @property
    def runtime_properties(self):
        return ctx.instance.runtime_properties

    @property
    def versions_repo_location(self):
        with open(ctx._endpoint.storage._payload_path) as f:
            return path(json.load(f)['versions_repo_location'])

    @versions_repo_location.setter
    def versions_repo_location(self, value):
        with ctx._endpoint.storage.payload() as payload:
            current_value = payload.get('versions_repo_location')
            if current_value and current_value != value:
                raise exceptions.NonRecoverableError(
                    'only one repository can be marked with "versions" type')
            payload['versions_repo_location'] = value

    @property
    def active_feature(self):
        with open(ctx._endpoint.storage._payload_path) as f:
            active_feature = json.load(f).get('active_feature')
        if not active_feature:
            return Feature({})
        branches = {}
        features_file = path(self.properties['features_file'])
        if features_file.exists():
            branches = yaml.safe_load(features_file.text()) or {}
        if not branches:
            return Feature({})
        name = active_feature
        active_feature = Feature(branches.get(active_feature))
        active_feature.name = name
        return active_feature

    def _git(self, log_out, repo_location=None):
        repo_location = repo_location or self.repo_location
        git = sh.git
        if log_out:
            git = bake(git)
        return git.bake(
            '--no-pager',
            '--git-dir', repo_location / '.git',
            '--work-tree', repo_location)

    def validate_active_feature(self):
        active_feature = self.active_feature
        current_branch = self.current_branch
        if not active_feature.branch:
            return False
        if active_feature.branch != current_branch:
            ctx.logger.warn(
                'Current branch: "{0}" does not match the active feature '
                'branch :{1}'.format(current_branch, active_feature))
            return False
        return True

    def _branch_exists(self, branch, local_only=False):
        branch_exists = self.git_output.bake(
                'show-ref', verify=True, quiet=True)
        refs = ['refs/heads']
        if not local_only:
            refs.append('refs/remotes/origin')
        for ref in refs:
            try:
                branch_exists('{0}/{1}'.format(ref, branch))
            except sh.ErrorReturnCode:
                pass
            else:
                return True
        return False

    def _fix_branch_name(self, branch):
        if not branch.startswith('.'):
            return branch
        if self.type not in ['core', 'plugin']:
            return branch
        template = '3{}' if self.type == 'core' else '1{}'
        return template.format(branch)

    def _read_versions_file(self, versions_branch):
        repo_git = self._git(
            log_out=False,
            repo_location=self.versions_repo_location)
        try:
            raw_versions = repo_git.show(
                '{}:versions.yaml'.format(versions_branch)).stdout.strip()
        except sh.ErrorReturnCode:
            raw_versions = repo_git.show(
                'origin/{}:versions.yaml'.format(
                    versions_branch)).stdout.strip()
        versions = yaml.safe_load(raw_versions)
        return versions.get('components', {})


repo = GitRepo()


class Hub(object):

    def ci_status(self):
        if not repo.validate_active_feature():
            return
        try:
            self.hub('ci-status', '-v').wait()
        except sh.ErrorReturnCode:
            pass

    def compare(self):
        if not repo.validate_active_feature():
            return
        url = self.hub_output.compare(
            '-u', '-b', repo.active_feature.base).stdout.strip()
        import webbrowser
        webbrowser.open(url)

    def pull_request(self, message, file):
        if not repo.validate_active_feature():
            return
        if message and file:
            raise exceptions.NonRecoverableError(
                'You cannot specify both a message and a file. Pick one!')
        command = self.hub.bake('pull-request', '-b', repo.active_feature.base,
                                '-o')
        if message:
            command = command.bake('-m', message)
        elif file:
            command = command.bake('-F', file)
        else:
            message = repo.active_feature.branch
            command = command.bake('-m', message)
        try:
            command()
        except sh.ErrorReturnCode:
            pass

    @property
    def hub(self):
        return self._hub(log_out=True)

    @property
    def hub_output(self):
        return self._hub(log_out=False)

    @staticmethod
    def _hub(log_out):
        repo_location = repo.repo_location
        try:
            hub = sh.hub
        except sh.CommandNotFound:
            raise exceptions.NonRecoverableError(
                'hub must be installed to use this command.')
        if log_out:
            hub = bake(hub)
        env = os.environ.copy()
        env.update({
            'GIT_WORK_TREE': repo_location,
            'GIT_DIR': repo_location / '.git'
        })
        return hub.bake(_env=env)


hub = Hub()


@workflow
def check_branch_exists(branch, **_):
    repos = []
    for instance in workflow_ctx.node_instances:
        if instance.node.type != 'git_repo':
            continue
        exists = instance.execute_operation('git.branch_exists', kwargs={
            'branch': branch
        }).get()
        if exists:
            repos.append(instance.node.id[:-len('-repo')])
    return repos


class Feature(dict):

    def __init__(self, initial):
        self.update(initial)

    @property
    def repos(self):
        return self.get('repos')

    @property
    def branch(self):
        repos = self.repos
        if not repos:
            return None
        elif isinstance(repos, dict):
            return repos.get(repo.name)
        elif isinstance(repos, list):
            if repo.name in repos:
                return self.get('branch')
            else:
                return None
        else:
            raise exceptions.NonRecoverableError('Invalid repos: {}'
                                                 .format(repos))

    @property
    def name(self):
        return self.get('name')

    @name.setter
    def name(self, value):
        self['name'] = value

    @property
    def base(self):
        return self.get('base', 'master')


def func(repo_method):
    @operation
    def wrapper(**kwargs):
        kwargs.pop('ctx', None)
        try:
            method = getattr(repo, repo_method)
        except AttributeError:
            method = getattr(hub, repo_method)
        return method(**kwargs)
    return wrapper


for method in ['clone', 'configure', 'pull', 'status', 'checkout', 'reset',
               'rebase', 'squash', 'diff', 'create_branch', 'branch_exists',
               'delete_branch', 'ci_status', 'compare', 'pull_request']:
    globals()[method] = func(method)



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

from path import path

from cloudify import ctx
from cloudify import exceptions
from cloudify.decorators import operation


@operation
def create_idea_project(virtualenv_name, **_):
    project_dir, repo_paths, package_paths, resource_dirs = _extract_dirs()
    if not project_dir:
        ctx.logger.info('No project dir configured.')
        return
    project_dir = path(project_dir)
    idea_dir = project_dir / '.idea'
    name_file = idea_dir / '.name'
    project_iml = idea_dir / '{}.iml'.format(project_dir.basename())
    misc_xml = idea_dir / 'misc.xml'
    modules_xml = idea_dir / 'modules.xml'
    vcs_xml = idea_dir / 'vcs.xml'
    module_paths = []
    for resource_dir in resource_dirs:
        if resource_dir.endswith('/'):
            resource_dir = resource_dir[:-1]
        resource_dir = path(resource_dir)
        iml_path = resource_dir / '{}.iml'.format(resource_dir.basename())
        module_paths.append(iml_path)
        if not iml_path.exists():
            ctx.logger.info('Adding resource module: {}'.format(iml_path))
            ctx.download_resource_and_render(
                resource_path='resources/idea/module.xml',
                target_path=iml_path,
                template_variables={'module_type': 'WEB_MODULE'})
    for package_path in package_paths:
        if package_path.endswith('/'):
            package_path = package_path[:-1]
        package_path = path(package_path)
        iml_path = package_path / '{}.iml'.format(package_path.basename())
        module_paths.append(iml_path)
        if not iml_path.exists():
            ctx.logger.info('Adding python module: {}'.format(iml_path))
            ctx.download_resource_and_render(
                resource_path='resources/idea/module.xml',
                target_path=iml_path,
                template_variables={'module_type': 'PYTHON_MODULE'})
    if not idea_dir.exists():
        idea_dir.mkdir_p()
    if not name_file.exists():
        name_file.write_text(virtualenv_name)
    if not project_iml.exists():
        ctx.download_resource_and_render(
            resource_path='resources/idea/module.xml',
            target_path=project_iml,
            template_variables={'module_type': 'JAVA_MODULE'})
    if not misc_xml.exists():
        ctx.download_resource_and_render(
            resource_path='resources/idea/misc.xml',
            target_path=misc_xml,
            template_variables={'virtualenv_name': virtualenv_name})
    if not modules_xml.exists():
        ctx.download_resource_and_render(
            resource_path='resources/idea/modules.xml',
            target_path=modules_xml,
            template_variables={'module_paths': module_paths})
    if not vcs_xml.exists():
        ctx.download_resource_and_render(
            resource_path='resources/idea/vcs.xml',
            target_path=vcs_xml,
            template_variables={'repo_paths': repo_paths})


def _extract_dirs():
    project_dir = None
    repo_paths = []
    package_paths = []
    resource_paths = []
    for rel in ctx.instance.relationships:
        node = rel.target.node
        instance = rel.target.instance
        props = node.properties
        runtime_props = instance.runtime_properties
        if node.type == 'git_repo':
            repo_location = runtime_props['repo_location']
            repo_paths.append(repo_location)
            if props['project_dir'] is True:
                if project_dir:
                    raise exceptions.NonRecoverableError(
                            'Cannot configure more than one project dir')
                project_dir = repo_location
            resources = props['resources'] or []
            if resources is True:
                resources = ['']
            repo_location = path(repo_location)
            for resource_path in resources:
                resource_paths.append(repo_location / resource_path)
        elif node.type == 'python_package':
            package_paths.append(runtime_props['package_path'])
    return project_dir, repo_paths, package_paths, resource_paths



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import argparse
import json
from contextlib import contextmanager

import argh
import colors
import yaml
from path import path

from clash import ctx


def ls():
    active_feature = features.active_feature
    for name, feature in features.load().items():
        _print_feature(name, feature, active_feature)


def _print_feature(name, feature, active_feature):
    indicator = '*' if name == active_feature else ' '
    print '{} {}:'.format(colors.red(indicator),
                          colors.magenta(name))
    indent = ' ' * 4
    branch = feature.get('branch')
    if branch:
        print '{}branch: {}'.format(indent, colors.green(branch))
    base = feature.get('base')
    if base:
        print '{}base: {}'.format(indent, colors.green(base))
    repos = feature.get('repos')
    if repos:
        print '{}repos:'.format(indent)
        if isinstance(repos, list):
            for repo in repos:
                print '{}- {}'.format(indent, colors.green(repo))
        else:
            for repo, branch in repos.items():
                print '{}  {}: {}'.format(indent,
                                          colors.green(repo),
                                          colors.green(branch))
    print


def create(name, branch, base):
    git_branch_exists = _command(workflow='check_branch_exists',
                                 parameters={'branch': branch})
    repos = sorted(_call(git_branch_exists))
    with features.update_feature(name) as feature:
        feature.update({
            'branch': branch,
            'repos': repos
        })
        if base:
            feature['base'] = base
    features.active_feature = name
    _print_feature(name, feature, active_feature=name)
    checkout(name)


def sync_repos(feature_name):
    feature_name = feature_name or features.active_feature
    if not feature_name:
        raise argh.CommandError('No feature is currently active.')
    if not features.exists(feature_name):
        raise argh.CommandError('No such feature: {}'.format(feature_name))
    branch = features.load()[feature_name]['branch']
    git_branch_exists = _command(workflow='check_branch_exists',
                                 parameters={'branch': branch})
    repos = sorted(_call(git_branch_exists))
    with features.update_feature(feature_name) as feature:
        feature.update({
            'repos': repos,
        })


def finish(feature_name):
    feature_name = feature_name or features.active_feature
    if not feature_name:
        raise argh.CommandError('No feature is currently active.')
    if not features.exists(feature_name):
        raise argh.CommandError('No such feature: {}'.format(feature_name))
    repos = features.load()[feature_name].get('repos', [])
    for repo in repos:
        try:
            remove_repo(repo, feature_name, force=False)
        except RuntimeError as e:
            print str(e), 'Skipping.'
    with features.update() as _features:
        _features.pop(feature_name, None)
    features.active_feature = None


def deactivate():
    features.active_feature = None
    _call(ctx.user_commands['git.checkout'], branch='default')


def checkout(name):
    feature = features.load().get(name)
    if feature:
        features.active_feature = name
    else:
        raise argh.CommandError('No such feature: {}'.format(name))
    _call(ctx.user_commands['git.checkout'], branch=name)


def add_repo(repo, feature_name):
    feature_name = feature_name or features.active_feature
    if not feature_name:
        raise argh.CommandError('No feature is currently active.')
    if not features.exists(feature_name):
        raise argh.CommandError('No such feature: {}'.format(feature_name))
    with features.update_feature(feature_name) as feature:
        branch = feature['branch']
        base = feature.get('base', 'master')
        repos = feature.get('repos', [])
        if repo not in repos:
            repos.append(repo)
        feature['repos'] = repos
    _call(_git_command(operation='git.create_branch',
                       repo=repo,
                       branch=branch,
                       base=base))
    _call(_git_command(operation='git.checkout',
                       repo=repo,
                       branch=feature_name))


def remove_repo(repo, feature_name, force):
    feature_name = feature_name or features.active_feature
    if not feature_name:
        raise argh.CommandError('No feature is currently active.')
    if not features.exists(feature_name):
        raise argh.CommandError('No such feature: {}'.format(feature_name))
    branch = features.load()[feature_name]['branch']
    _call(_git_command(operation='git.delete_branch',
                       repo=repo,
                       branch=branch,
                       force=force))
    with features.update_feature(feature_name) as feature:
        repos = feature.get('repos', [])
        if repo in repos:
            repos.remove(repo)
        feature['repos'] = repos


class Features(object):

    @property
    def path(self):
        return path(ctx.user_config.inputs['features_file'])

    def load(self):
        if not self.path.exists():
            return {}
        return yaml.safe_load(self.path.text()) or {}

    def save(self, value):
        self.path.write_text(yaml.safe_dump(value, default_flow_style=False))

    @contextmanager
    def update(self):
        content = self.load()
        yield content
        self.save(content)

    def exists(self, name):
        return name in self.load()

    @contextmanager
    def update_feature(self, name):
        with self.update() as content:
            feature = content.get(name, {})
            yield feature
            content[name] = feature

    @property
    def active_feature(self):
        with open(ctx.env.storage._payload_path) as f:
            return json.load(f).get('active_feature')

    @active_feature.setter
    def active_feature(self, value):
        with ctx.env.storage.payload() as payload:
            payload['active_feature'] = value


features = Features()


def _git_command(operation, repo, **kwargs):
    return _command(
        workflow='execute_operation',
        parameters={
            'operation': operation,
            'operation_kwargs': kwargs,
            'type_names': ['git_repo'],
            'node_ids': ['{}-repo'.format(repo)]
        },
        event_cls='clue.output:NamedNodeEvent')


def _command(workflow, parameters=None, event_cls=None):
    parameters = parameters or {}
    command = {
        'workflow': workflow,
        'parameters': parameters,
    }
    if event_cls:
        command['event_cls'] = event_cls
    return ctx._parse_command('generated', command)


def _call(command, **kwargs):
    if 'verbose' not in kwargs:
        kwargs['verbose'] = False
    return command(argparse.Namespace(**kwargs))



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

import colors

from clash.output import Event


class NamedNodeEvent(object):

    @staticmethod
    def factory(env, verbose, command):
        max_len = 0
        git_command = command.get('parameters', {}).get(
            'operation').startswith('git.')
        node_type = 'git_repo' if git_command else 'python_package'
        for node in env.storage.get_nodes():
            if node.type == node_type:
                max_len = max(max_len, len(node.properties['name']))
        formatting = '{0:<' + str(max_len + 1) + '}'

        class NamedNodeEventImpl(Event):
            def __str__(self):
                if (not verbose and self.level and
                        self.level.lower() in ['info', 'warning'] and
                        self.node_name):
                    node = env.storage.get_node(self.node_name)
                    name = node.properties['name']
                    name = colors.green(formatting.format(name))
                    return ' {0}| {1}'.format(name, self.message)
                else:
                    return super(NamedNodeEventImpl, self).__str__()
        return NamedNodeEventImpl


class NoseEvent(object):

    @staticmethod
    def factory(env, verbose, command):
        class NoseEventImpl(Event):
            def __str__(self):
                if not verbose:
                    return self.message
                else:
                    return super(NoseEventImpl, self).__str__()
        return NoseEventImpl



########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

from setuptools import setup

setup(
    name='clue',
    version='0.38',
    author='GigaSpaces',
    author_email='cosmo-admin@gigaspaces.com',
    packages=['clue'],
    description='Cloudify development environment CLI',
    license='Apache License, Version 2.0',
    zip_safe=False,
    install_requires=[
        'sh',
        'virtualenvwrapper',
        'clash==0.17'
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'clue = clue:main',
        ]
    }
)
